### stuff that adds dimensionality reductions when called on a scanpy object
import scanpy as sc
from sklearn.neighbors import NearestNeighbors
from .score import info_score
import numpy as np
from scipy.sparse import csr_matrix
import warnings

def reduce(X, n_comps = 50, iters=1, max_bins=float('inf'), fast_version=True, scaled_bins=False, nbhds=None,
           rep=None, nbhd_size=15, n_pcs=50, metric='cosine',
           keep_scores=False, keep_loadings=False, keep_all_iters=False, verbose=False, **kwargs):


    """
    :param X: (num cells)x(num_genes)-sized array or sparse matrix to be dimensionality-reduced. Should be nonnegative,
    with 0 indicating no recorded transcripts (required for binarization and binomial inference).
    :param n_comps: Desired dimensionality of the reduction.
    :param iters: Number of iterations. More iterations yi
    :param max_bins: Resolution of gene frequency measurements. If max_bins=inf (default), each global gene frequency is
    measured exactly. Otherwise, they are binned into max_bins bins, reducing the number of binomial tests required.
    :param fast_version: If true (default), runtime will be faster but memory usage will be higher. If false, will be slower with
    a smaller memory footprint.
    :param scaled_bins: If true, frequency bins will be unevenly spaced over the unit interval via a power law,
    ensuring accuracy to a constant factor. If false (default), bins evenly divide the unit interval.
    :param nbhd_size: Size of neighborhoods used to assess the local expression of a gene. Should be smaller than the
    smallest subpopulation; default is 15. Does not drastically affect the
    :param nbhds: Optional - if k-neighborhoods of points are already determined, they can be specified here as
    a (num_cells)*k array. Otherwise, they will be computed from the PCS.
    :param rep: Optional - existing featurization for determining k-neighborhoods of points. If not provided,
    k-neighborhoods will be computed in PCA space.
    :param n_pcs: If either rep or nbhds is provided, this parameter is ignored. Otherwise, the number of PCs used for computing
    the starting representation, from which the starting k-neighborhoods are derived.
    :param metric: if either rep or nbhds is provided, this parameter is ignored. Otherwise, the metric used in computing
    the initial k-neighborhoods
    :param keep_scores: if True, returns information scores for each gene/cell combo in a sparse matrix.
    :param keep_loadings: If True, returns loadings of each gene in each metagene as a dense matrix.
    :param verbose: If True, print progress.
    :param kwargs: Other arguments to be passed to info_score
    :return: If return_scores or return_loadings are both false, a (n cells)x(n_comps)-dimensional array
    of reduced features. Otherwise, a dictionary with keys 'reduction', 'scores' and/or 'loadings'.
    """

    result = {} # final result, if return_all_iters=True

    if nbhds is None:
        if rep is None:
            rep = sc.tl.pca(X, n_comps=n_pcs)

        nn = NearestNeighbors(n_neighbors=nbhd_size, metric=metric)
        nn.fit(rep)

        nbhds = nn.kneighbors(return_distance=False)

        # add points to their own neighborhoods
        nbhds = np.concatenate((np.array(range(X.shape[0])).reshape(-1, 1), nbhds), axis=1)


    for i in range(iters):
        if verbose:
            print('\niteration {}'.format(i+1))

        if i==0:
            #keep binomial scores for future runs
            infos, gene_bins, binom_scores = info_score(X, nbhds,
                                                        max_bins=max_bins,
                                                        return_bin_info=True,
                                                        fast_version=fast_version,
                                                        verbose=verbose,
                                                        scaled=scaled_bins,
                                                        **kwargs)
        else:
            infos = info_score(X, nbhds,
                               max_bins=max_bins,
                               binom_scores=binom_scores,
                               gene_bins=gene_bins,
                               fast_version=fast_version,
                               verbose=verbose,
                               scaled=scaled_bins,
                               **kwargs)


        a, bt, c, d = sc.tl.pca(infos, n_comps=n_comps, return_info=True)

            # pca(infos, k=n_comps, raw=True, pc_iters=pc_iters)

        current_dimred = X @ bt.transpose() # metagene expression values in X

        if keep_all_iters:
            result['reduction_{}'.format(i+1)] = current_dimred

        # compute neighborhoods for next iteration, using current reduction
        if i < (iters-1):
            nn = NearestNeighbors(n_neighbors=nbhd_size, metric=metric)
            nn.fit(current_dimred)

            nbhds = nn.kneighbors(return_distance=False)
            nbhds = np.concatenate((np.array(range(X.shape[0])).reshape(-1, 1), nbhds), axis=1)


    if not keep_scores and not keep_loadings and not keep_all_iters:
        return current_dimred


    if not keep_all_iters:
        result['reduction'] =  current_dimred #store result
    if keep_scores:
        result['scores'] = csr_matrix(infos)
    if keep_loadings:
        result['loadings'] = bt.transpose()

    return result


def reduce_scanpy(adata, keep_scores=False, keep_loadings=True, keep_all_iters=False, layer=None, key_added='sca', iters=1, **kwargs):
    if layer is not None:
        X = adata.layers[layer]
    else:
        X = adata.X

    dimred_info = reduce(X, keep_scores=keep_scores, keep_loadings=keep_loadings,
                           keep_all_iters=keep_all_iters, iters=iters, **kwargs)

    if not keep_scores and not keep_loadings and not keep_all_iters:
        adata.obsm[key_added] = dimred_info
    else:
        if keep_all_iters:
            for i in range(iters):
                adata.obsm[key_added+'_'+str(i+1)] = dimred_info['reduction_{}'.format(i+1)]
        else:
            adata.obsm[key_added] = dimred_info['reduction']
            
        if 'scores' in dimred_info:
            adata.layers[key_added+'_score'] = dimred_info['scores']
        if 'loadings' in dimred_info:
            adata.varm[key_added+'_loadings'] = dimred_info['loadings']


###### DEPRECATED #########

def info_pca(nbhd_size=15, n_neighbors=15, n_init_pcs=50, n_info_pcs=50, key_added='info_pca', metric='cosine',
             max_bins=float('inf'), iters=1, entropy_normalize=False, p_val=True,
             keep_scores=True, binarize=False,  verbose=False, **kwargs):

    warnings.warn('This function has been replaced by reduce and reduce_scanpy, and will soon be removed.' )

    def f(d):
        #sc.tl.pca(d, n_comps=n_init_pcs)

        if binarize:
            pcs = sc.tl.pca((d.X>0).astype('float'), n_comps = n_init_pcs)
            d.obsm['X_pca'] = pcs
            # U,s,Vt = pca((d.X>0).astype('float'), k=n_init_pcs, n_iter=pc_iters)
            # pcs = U*s
        else:
            # U,s,Vt = pca(d.X, k=n_init_pcs, n_iter=5)
            # pcs = U*s
            # d.obsm['X_pca'] = pcs
            sc.tl.pca(d, n_comps=n_init_pcs)
            pcs = d.obsm['X_pca']


        nn = NearestNeighbors(n_neighbors=nbhd_size, metric=metric);
        nn.fit(pcs)
        nbhds = nn.kneighbors(return_distance=False)

        # add points to their own neighborhoods
        nbhds = np.concatenate((np.array(range(d.shape[0])).reshape(-1, 1), nbhds), axis=1)


        for i in range(iters):
            if verbose:
                print('\niteration {}'.format(i))

            if i==0:
                #keep binomial scores for future runs
                infos, gene_bins, binom_scores = info_score((d.X>0).astype('float'), nbhds,
                                                            max_bins=max_bins, entropy_normalize=entropy_normalize,
                                                            return_bin_info=True, **kwargs)
            else:
                infos = info_score((d.X>0).astype('float'), nbhds, max_bins=max_bins,
                                   binom_scores = binom_scores, gene_bins=gene_bins, **kwargs)

            a, bt, c, dt = sc.tl.pca(infos, n_comps=n_info_pcs, return_info=True)
            #U, s, Vt = pca(infos, k=n_info_pcs, raw=True)

            if binarize:
                X = (d.X>0).astype('float').todense()

                d.obsm[key_added] = X @ bt.transpose()
                    #np.matmul(X, Vt.transpose())
            else:
                d.obsm[key_added] = d.X @ bt.transpose()

            d.varm[key_added+'_loadings'] = bt.transpose()

            #print(d.obsm[key_added].shape)

            nn = NearestNeighbors(n_neighbors=nbhd_size);
            nn.fit(d.obsm[key_added])

            dist, nbhds = nn.kneighbors()
            nbhds = np.concatenate((np.array(range(d.shape[0])).reshape(-1, 1), nbhds), axis=1)

        sc.pp.neighbors(d, use_rep=key_added, key_added=key_added, metric=metric, n_neighbors=n_neighbors)

        if keep_scores:
            # info_df = pd.DataFrame(infos)
            # info_df.columns = d.var.index
            # info_df.index = d.obs.index
            d.layers[key_added+'_score'] = csr_matrix(infos)

    return f


import scanpy as sc
import numpy as np
import matplotlib.pyplot as plt
from .utils import metagene_loadings
import pandas as pd
import seaborn as sns

def ordered_matrixplot(d, n_genes=5, groups=None, **kwargs):
    """
    matrix plot of ranked groups, with columns ordered by score instead of abs(score).
    Separates up- from down-regulated genes better, resulting in visually-cleaner plots
    :param d:
    :param n_genes:
    :param kwargs:
    :return:
    """
    top_genes = np.stack([np.array(list(x)) for x in d.uns['rank_genes_groups']['names']][:n_genes])
    top_scores = np.stack([np.array(list(x)) for x in d.uns['rank_genes_groups']['scores']][:n_genes])

    # order top genes by actual score, not absolute value
    ordered_top_genes = np.take_along_axis(top_genes, np.argsort(-1 * top_scores, axis=0), axis=0)


    # print(ordered_top_genes)



    grouping_key = d.uns['rank_genes_groups']['params']['groupby']
    group_names = list(d.uns['rank_genes_groups']['names'].dtype.fields.keys())

    ordered_top_mapping = {group_names[i]: ordered_top_genes[:, i] for i in range(len(group_names))}


    if groups is not None:
        ordered_top_mapping = {k:v for k, v in ordered_top_mapping.items() if k in groups}
        #print(ordered_top_mapping)


    sc.pl.matrixplot(d, var_names=ordered_top_mapping, groupby=grouping_key, **kwargs)

def plot_metagenes(data, comps=None, key='sca', **kwargs):
    if comps is None:
        if type(data) is dict:
            comps = list(range(data['loadings'].shape[1]))
        else:
            comps = list(range(data.varm[key+'_loadings'].shape[1]))
    fig, axs = plt.subplots(len(comps),1)

    loadings = metagene_loadings(data, key=key, **kwargs)

    for i,comp in enumerate(comps):
        df = pd.DataFrame(loadings[comp])
        sns.barplot(data=df, x='genes', y='scores', ax = axs.flatten()[i])
        axs.flatten()[i].set_title('component {}'.format(i+1))

        for tick in axs.flatten()[i].get_xticklabels():
            tick.set_rotation(45)

    fig.set_size_inches( 0.5*df.shape[0], 5*len(comps))
    plt.subplots_adjust(hspace=0.5)


    return fig





import numpy as np


def metagene_loadings(data, n_genes=10, rankby_abs=False, key='sca'):
    """
    return the top-loaded genes for each Scalpel metagene
    :param data: AnnData or dict output from reduce_scanpy or reduce. Must contain loading data.
    :param n_genes: Number of top-loaded genes to record
    :param rankby_abs: If True, record the top genes by absolute value.
    :param key: Namespace of Scalpel data, if data is a scanpy object. Will look under data.varm[key+'_loadings'].
    :return: Dictionary mapping [sca component number]:{[top gene 1]:loading, [top_gene_2]:loading... [top_gene_k]:loading}
    """

    if type(data) is dict:
        loadings = data['loadings']
        var_names = np.array(list(range(loadings.shape[0])))
    else:
        loadings = data.varm[key+'_loadings']
        var_names = np.array(data.var_names)

    loadings /= np.sum(np.abs(loadings), axis=0) #normalize to percent contribution
    loadings *= loadings.shape[0] # fold enrichment over expected

    if rankby_abs:
        idxs = np.transpose(np.argsort(np.abs(loadings), axis=0)[(-1*n_genes):,:]).tolist()
    else:
        idxs = np.transpose(np.argsort(loadings, axis=0)[(-1 * n_genes):,:]).tolist()

    genes = [var_names[np.array(x)] for x in idxs]
    scores = [loadings[x,i] for i, x in enumerate(idxs)]


    result = list(zip(genes, scores))
    result = {i:{'genes':k[np.argsort(-1*v)], 'scores':v[np.argsort(-1*v)]} for i,(k,v) in enumerate(result)}
    return result



from scipy.stats import binom_test, binom, entropy
import numpy as np
from scipy.sparse import csr_matrix


def taylor_exp(x, n):
    # taylor series of 1-(1-x)^n
    # use for multiple test correction if x is very small.
    return (n*x-
           0.5*np.power(x,2)*(n-1)*n+
           (1./6.)*np.power(x,3) * (n-2)*(n-1)*n-
           (1./24)*np.power(x,4)*(n-3)*(n-2)*(n-1)*n)

def get_binom_scores(gene_probs, k, max_bins=500,  verbose=True, scaled=False, n_tests=50, multi_correct=True, error_rate=1.0):
    # compute binom_test(x, k, p) for all x in 1:k and all p in gene_probs
    # if scaled, put more density in lower probabilities
    if max_bins < float('inf'):
        # split interval into max_bins bins

        if scaled:
            if verbose:
                print('using scaled bins')
            min_prob = np.min(gene_probs[gene_probs>0])
            precision = (min_prob)**(1./float(max_bins)) # fold accuracy in probability approx
            splits = precision ** (max_bins-(np.arange(max_bins)))
        else:
            splits = np.arange(0, 1, 1/max_bins)[1:] # don't allow a gene to have 0 prob...

        #bins each gene gets put in, based on its global probability
        gene_bins = np.array([np.argmin(np.abs(gene_probs[j] - splits)) for j in range(len(gene_probs))])

        binom_scores = np.zeros((len(splits), k+1))

        for i, p in enumerate(splits):
            if verbose:
                print('\r computing binom scores for bin {}/{}'.format(i, len(splits)), end=' ')

            # #print('using fast scores')
            # if fast_scores:
            signs = [1 if p*k <= j else -1 for j in np.arange(k+1)]
            pmfs = binom.pmf(np.arange(k+1), k, p)
            orderer = np.argsort(pmfs)

            pvals = np.cumsum(pmfs[orderer])





            if multi_correct:
                # use FWER to correct for testing many genes
                pvals_corrected = 1-np.power(1-pvals, n_tests)
                pvals_corrected[pvals<1e-10] = taylor_exp(pvals[pvals<1e-10], n_tests) # more accurate

                pvals = pvals_corrected

            pvals[pvals > error_rate] = 1.  # p<error rate or it didn't happen
            binom_scores[i,orderer] = np.array(signs)[orderer] * -1*np.log(pvals)

            # else:
            #     for j in range(k+1):
            #         alt = 'two-sided' if two_tailed else 'less' if float(j)/k < p*k else 'greater'
            #
            #         sign = 1 if p*k <= j else -1  # negative if lower than expected
            #
            #         binom_scores[i,j] = sign * -1*np.log(binom_test(j, n=k, p=p, alternative=alt))


    else:  # compute for all gene probs
        binom_scores = np.zeros((len(gene_probs), k + 1))
        gene_bins = np.array(range(len(gene_probs)))

        for i in range(len(gene_probs)):
            p=gene_probs[i]

            if verbose:
                print('\r computing binom scores for genes {}/{}'.format(i, len(gene_probs)), end=' ')
            if gene_probs[i] == 0 or gene_probs[i] == 1: # no expression, no score
                continue

            signs = [1 if p*k <= j else -1 for j in np.arange(k+1)]
            pmfs = binom.pmf(np.arange(k+1), k, p)
            orderer = np.argsort(pmfs)

            pvals = np.cumsum(pmfs[orderer])


            if multi_correct:
                # use FWER to correct for testing many genes
                pvals_corrected = 1-np.power(1-pvals, n_tests)
                pvals_corrected[pvals<1e-10] = taylor_exp(pvals[pvals<1e-10], n_tests) # more accurate

                pvals = pvals_corrected
               #pvals = 1-np.power(1-pvals, n_tests)

            pvals[pvals > error_rate] = 1.  # p<error rate or it didn't happen

            binom_scores[i,orderer] = np.array(signs)[orderer] * -1*np.log(pvals)
            #
            # for j in range(k + 1):
            #
            #
            #
            #     p = gene_probs[i]
            #     alt = 'two-sided' if two_tailed else 'less' if float(j)/k < p*k else 'greater'
            #
            #     sign = 1 if p*k <= j else -1  # negative if lower than expected
            #
            #     binom_scores[i,j] = sign * -1*np.log(binom_test(j, n=k, p=p, alternative=alt))

    return (gene_bins, binom_scores)

def info_score(X, nbhds, max_bins=float('inf'),
               entropy_normalize=False, fast_version=True, binom_scores=None, gene_bins=None,
               return_bin_info=False, verbose=True, n_tests = 50, **kwargs):
    """
    :param X: sparse count matrix
    :param nbhds: list with indices of nearest neighbors for each obs in X, e.g. from kneighbors() in sklearn
    :param max_bins: Resolution at which global gene probabilities are computed.
    if inf, genes get their own probabilities. Otherwise, the unit interval is split into max_bins pieces
    and they are rounded. This makes it faster with little performance difference
    :param return_all: if True, will also return global and local gene probabilities

    :param binom_scores: pass in binomial scores for each gene/bin, if pre-computed. Allows saving for future iterations.
    :param gene_bins: pass in gene bins from previous run. Speeds up iteration

    :param return_bin_info: for iteration: keep information about gene bins and binomial probs.
    :param fast_version: if True, use matrix multiplication instead of iteration. Fast, but memory-intensive.
    :return: dense matrix of gene/cell weightings.
    """

    X = csr_matrix((X>0).astype('float'))  # convert to sparse binarized matrix

    if type(nbhds) is np.ndarray:
        nbhds = list(nbhds)

    k = len(nbhds[0]) # how many neighbors?

    wts = np.zeros(X.shape)  # too big for large data
    # nbhd_counts = np.zeros(X.shape)  # ditto
    # nbhd_sizes = [len(x) for x in NNs]

    # first compute frequencies of all genes:
    gene_probs = np.array((X > 0).sum(axis=0) / float(X.shape[0])).flatten()

    #frequencies of genes within neighborhoods
    nbhd_probs = np.zeros(X.shape)


    if binom_scores is None or gene_bins is None:
        if n_tests is None:
            n_tests = X.shape[1] # multi-correct per cell
        gene_bins, binom_scores = get_binom_scores(gene_probs, k, max_bins=max_bins,
                                                   verbose=verbose, n_tests = n_tests, **kwargs)




    if fast_version:
        # compute significance of gene expression in each cell's neighborhood
        #first convert neighborhood to sparse matrix
        data = np.ones(np.sum([len(x) for x in nbhds]))
        col_ind = [item for sublist in nbhds for item in sublist]
        row_ind = [i for i,sublist in enumerate(nbhds) for item in sublist]

        #sparse adjacency matrix of NN graph
        nn_matrix = csr_matrix((data, (row_ind, col_ind)), shape=(X.shape[0], X.shape[0]))

        # get gene expressions within each neighborhood; this matrix may be less sparse
        nbhd_exprs = (nn_matrix * X).astype('int').todense()

        # # extract locations and values of nonzero nbhd expressions.
        # rows, cols = nbhd_exprs.nonzero()
        # exprs = nbhd_exprs.data

        #apply binomial scores
        rows, cols = np.indices((X.shape))
        rows = rows.flatten()
        cols = cols.flatten()

        wts = binom_scores[gene_bins[cols], np.array(nbhd_exprs[rows, cols]).flatten()].reshape(X.shape)

    else:
        for i in range(X.shape[0]):

            if verbose:
                if i < X.shape[0]-1:
                    print('\r computing counts for cell {}/{}'.format(i, X.shape[0]), end='         ')
                else: print('\r computing counts for cell {}/{}'.format(i, X.shape[0]), end='         \n')

            nnbhd = X[nbhds[i], :]

            nbhd_size = len(nbhds[i])
            nbhd_gene_counts = np.array((nnbhd > 0).sum(axis=0)).flatten()

            nbhd_probs[i,:] = nbhd_gene_counts/nbhd_size

            if max_bins < float('inf'):
                # look up the binomial score in the nearest bins

                # gene_scores = [binom_scores[gene_bins[j], count] for j, count in enumerate(nbhd_gene_counts)]
                gene_scores = binom_scores[gene_bins, nbhd_gene_counts]

            else:
                gene_scores = [binom_scores[j, count] for j, count in enumerate(nbhd_gene_counts)]

            wts[i,:] = gene_scores
            # expected_vals = nbhd_size * gene_probs
            # wts[i, :] = -1*np.log(gene_scores) * (2*(nbhd_gene_counts > expected_vals)-1)





    if entropy_normalize: # divide each column by the entropy of the corresponding gene
        gene_entropies = -1*(np.multiply(gene_probs,np.log(gene_probs))+
                             np.multiply((1-gene_probs),np.log(1-gene_probs)))
        gene_entropies[np.logical_not(np.isfinite(gene_entropies))] = float('inf') # zeros out non-expressed or everywhere-expressed genes
        wts = np.divide(wts, gene_entropies)

    # if return_all:
    #     return (wts, gene_probs, nbhd_probs)

    if return_bin_info: # for iteration
        return (wts, gene_bins, binom_scores)
    else:
        return (wts)





    #
    # # pre-compute binomial scores to avoid doing it for each cell/gene
    # if max_bins < float('inf') and binom_scores is None:
    #     # split interval into max_bins bins
    #     splits = np.arange(0, 1, 1/max_bins)[1:] # don't allow a gene to have 0 prob...
    #     gene_bins = np.array([np.argmin(np.abs(gene_probs[j] - splits)) for j in range(X.shape[1])])
    #
    #

    #     binom_scores = np.zeros((len(splits), k+1))
    #     for i, p in enumerate(splits):
    #         print('\r computing binom scores for bin {}/{}'.format(i, len(splits)), end=' ')
    #         for j in range(k+1):
    #             alt = 'two-sided' if two_tailed else 'less' if float(j)/k < p*k else 'greater'
    #
    #             if p_val: # information of any occurrence more extreme
    #                 binom_scores[i,j] = binom_test(j, n=k, p=p, alternative=alt)
    #             else: #information of this actual occurrence
    #                 val = -1*(j*np.log(p)+(k-j)*np.log(1-p))
    #                 if np.isfinite(val): # zero out nans
    #                     binom_scores[i,j] = -1*(j*np.log(p)+(k-j)*np.log(1-p))

    #
    # elif binom_scores is None: # compute for all gene probs
    #     binom_scores = np.zeros((X.shape[1], k + 1))
    #     for i in range(X.shape[1]):
    #         print('\r computing binom scores for genes {}/{}'.format(i, X.shape[1]), end=' ')
    #         for j in range(k + 1):
    #             p=gene_probs[i]
    #             alt = 'two-sided' if two_tailed else 'less' if float(j)/k < p*k else 'greater'
    #             if p_val: # information of any occurrence more extreme
    #                 binom_scores[i,j] = binom_test(j, n=k, p=p, alternative=alt)
    #             else: #information of this actual occurrence
    #                 val = -1*(j*np.log(p)+(k-j)*np.log(1-p))
    #                 if np.isfinite(val): # zero out nans
    #                     binom_scores[i,j] = -1*(j*np.log(p)+(k-j)*np.log(1-p))
